# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Model base"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import weakref

from collections import OrderedDict, namedtuple
from functools import wraps

from future.utils import with_metaclass, viewitems, viewvalues, viewkeys
from sqlalchemy import Column
from sqlalchemy.orm import relationship

from .. import relational


class MetaModel(type):
    """Model metaclass

    Keep track of all model classes and instances
    Classes must have __modelname__ defined to be tracked
    """
    __classes__ = {}

    def __new__(mcs, name, bases, attrs):
        attrs["__refs__"] = []
        attrs["REPLACE"] = attrs.get("REPLACE", {})
        attrs["DEFAULT"] = attrs.get("DEFAULT", {})
        if "__modelname__" not in attrs:
            attrs["__modelname__"] = name

        cls = super(MetaModel, mcs).__new__(mcs, name, bases, attrs)
        mcs.__classes__[attrs["__modelname__"]] = cls
        return cls

    def __call__(cls, *args, **kwargs):
        instance = super(MetaModel, cls).__call__(*args, **kwargs)
        instance.__class__.__refs__.append(weakref.ref(instance))
        return instance

    def get_instances(cls):
        """Return all instances from Model class"""
        for inst_ref in cls.__refs__:
            inst = inst_ref()
            if inst is not None:
                yield inst

    def set_instances_default(cls, attr, value):
        """Set DEFAULT attribute for instances of classes created
        by this metaclass


        Arguments:
        attr -- attribute name
        value -- new attribute value
        """
        for instance in cls.get_instances():
            instance.set_instance_attr(attr, value)

    @classmethod
    def all_models(mcs):
        """Return all instances from all models"""
        for cls in viewvalues(mcs.__classes__):
            for instance in cls.get_instances():
                yield instance

    @classmethod
    def set_class_default(mcs, model, attr, value, instances=False):
        """Set DEFAULT attribute for Model class

        Arguments:
        model -- name of model class
        attr -- attribute name
        value -- new attribute value


        Keyword arguments:
        instances -- update instances too (default=False)
        """
        cls = mcs.__classes__[model]
        if attr in cls.REPLACE:
            attr = cls.REPLACE[attr]
        if attr in cls.DEFAULT:
            cls.DEFAULT[attr] = value
        if instances:
            cls.set_instances_default(attr, value)

    @classmethod
    def set_classes_default(mcs, attr, value, instances=False, model="*"):
        """Set DEFAULT attribute for Model classes that match model filter

        Arguments:
        model -- name of model class
        attr -- attribute name
        value -- new attribute value


        Keyword arguments:
        instances -- update instances too (default=False)
        model -- filter model (default="*")
        """
        if model == "*":
            for name in viewkeys(mcs.__classes__):
                mcs.set_class_default(name, attr, value, instances=instances)
        else:
            mcs.set_class_default(model, attr, value, instances=instances)


class Model(with_metaclass(MetaModel)):
    """Model base"""

    REPLACE = {}
    DEFAULT = {}

    def __init__(self, *args, **kwargs):
        pass

    def set_instance_attr(self, attr, value, check=True):
        """Initialize attr to instance

        If attr exists in REPLACE, use its variant
        If attr has ".", use getattr until the actual item
        """
        if attr in self.REPLACE:
            attr = self.REPLACE[attr]
        obj = self
        while "." in attr:
            attr0, attr = attr.split(".", 1)
            obj = getattr(self, attr0)
        if not check or hasattr(obj, attr):
            setattr(obj, attr, value)

    def initialize_default(self, kwargs):
        """Initialize DEFAULT and kwargs parameters to instance"""
        for key, value in viewitems(self.DEFAULT):
            self.set_instance_attr(key, value, check=False)
        for key, value in viewitems(kwargs):
            self.set_instance_attr(key, value)


def proxy(element):
    """Return proxy instance from SQLALchemy object"""
    cls = element.__class__
    if cls in AlchemyProxy.__alchemy_refs__:
        return AlchemyProxy.__alchemy_refs__[cls](element)
    else:
        return element


def proxy_gen(query):
    """Return proxy generator from iterable

    Iterable can be a SQLALchemy query"""
    for element in query:
        yield proxy(element)


def proxy_property(func, proxy_func=proxy):
    """Return a proxy property to a __model__ function"""
    @wraps(func)
    def prop(self, *args, **kwargs):
        """Proxy property"""
        obj = self._get_instance()                                               # pylint: disable=protected-access
        result = func(obj, *args, **kwargs)
        return proxy_func(result)
    return property(prop)


def proxy_attr(name, proxy_func=proxy):
    """Return a proxy property to a __model__ attribute"""
    def func(self):
        """Return {}""".format(name)
        return getattr(self, name)
    return proxy_property(func, proxy_func=proxy_func)


class AlchemyProxy(Model):
    """Alchemy Proxy super class.

    Store _alchemy_pk primary key and other columns from table
    """

    __alchemy_refs__ = {}

    m = __model__ = None                                                         # pylint: disable=invalid-name
    t = __table__ = None                                                         # pylint: disable=invalid-name
    __modelname__, __columns__ = None, []

    def __init__(self, obj):
        super(AlchemyProxy, self).__init__(obj)
        if isinstance(obj, relational.base):
            self._store_pk(obj)
        else:
            self._alchemy_pk = obj
        self._restore_instance()

    def _store_pk(self, obj):
        self._alchemy_pk = obj.__mapper__.primary_key_from_instance(obj)

    def _restore_instance(self):
        """Restore instance with new session"""
        obj = self._get_instance()
        for column in self.__columns__:
            setattr(self, column, getattr(obj, column))

    def _get_instance(self):
        return relational.session.query(self.__model__).get(self._alchemy_pk)

    def __getstate__(self):
        return (self._alchemy_pk,)

    def __setstate__(self, state):
        (self._alchemy_pk,) = state
        self._restore_instance()

    def to_dict(self, ignore=tuple(), extra=tuple()):
        """Return object as dict"""
        result = OrderedDict(
            (attr, getattr(self, attr)) for attr in extra
        )
        for key in self.__columns__:
            if key not in ignore and key not in extra:
                result[key] = getattr(self, key)

        return result

    @classmethod
    def fast_store(cls, trial_id, object_store, partial, conn=None):
        """Bulk insert lightweight objects from ObjectStore"""
        if object_store.has_items():
            _conn = conn if conn else relational.engine.connect()
            _conn.execute(
                cls.__model__.__table__.insert().prefix_with("OR REPLACE"),
                *object_store.generator(trial_id, partial)
            )
            if conn is None:
                _conn.close()

def create_relationship(proxy_func):
    """Create proxy descriptor"""
    class Relationship(object):                                                  # pylint: disable=too-few-public-methods
        """Create a proxy for relationship
        Relationship on Model class will be prepended by _
        """
        def __init__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs
            self.name = None

        def __get__(self, obj, objtype=None):
            if obj is None:
                return self
            alchemy = obj._get_instance()                                        # pylint: disable=protected-access
            return proxy_func(getattr(alchemy, self.name))
    return Relationship


ModelMethod = namedtuple("ModelMethod", "func proxy")


def query_many_property(func):
    """Property is part of the Model class. It should return a generator"""
    return ModelMethod(func, proxy_gen)


Many = create_relationship(proxy_gen)
One = create_relationship(proxy)


def one(*args, **kwargs):
    """Create One relationship"""
    return One(*args, **kwargs)


def many(*args, **kwargs):
    """Create Many relationship"""
    if "lazy" not in kwargs:
        kwargs["lazy"] = "dynamic"
    return Many(*args, **kwargs)


def many_ref(backref, *args, **kwargs):
    """Create Many relationship with backref"""
    kwargs["backref"] = backref
    return many(*args, **kwargs)


def many_viewonly_ref(backref, *args, **kwargs):
    """Create Many relationship with backref and viewonly"""
    kwargs["backref"] = backref
    kwargs["viewonly"] = True
    return many(*args, **kwargs)


def backref_many(name):
    """Create property for backref generator"""
    return proxy_attr(name, proxy_func=proxy_gen)


def backref_one(name):
    """Create property for backref object"""
    return proxy_attr(name)


def proxy_class(cls):
    """Proxy decorator


    Use it for classes that define SQLAlchemy attributes
    This decorator creates __model__, __table__, __columns__ attributes
    It will also register the class in the proxy list
    """
    description = cls.__dict__
    attributes = {}
    to_remove = set()
    for name, var in viewitems(description):
        if isinstance(var, Column):
            to_remove.add(name)
            #description[name] = None
            attributes[name] = var
        elif isinstance(var, (Many, One)):
            var.name = name
            attributes[var.name] = relationship(*var.args, **var.kwargs)
        elif isinstance(var, ModelMethod):
            new_name = name
            setattr(cls, name, proxy_attr(new_name, proxy_func=var.proxy))
            attributes[new_name] = property(var.func)
        elif name in ('__tablename__', '__table_args__'):
            attributes[name] = var

    for name in to_remove:
        delattr(cls, name)

    cls.__modelname__ = cls.__name__
    cls.m = cls.__model__ = type(cls.__name__, (relational.base,), attributes)
    cls.t = cls.__table__ = cls.__model__.__table__
    cls.__columns__ = cls.__table__.columns.keys()

    AlchemyProxy.__alchemy_refs__[cls.__model__] = cls

    return cls


def is_none(attribute):
    """Filter SQLAlchemy attribute is None"""
    return attribute == None                                                     # pylint: disable=singleton-comparison



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""IPython Magics"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import sys

from IPython.utils.process import arg_split
from IPython.core import magic_arguments
from IPython.core.magic import line_magic, cell_magic, line_cell_magic

from ...cmd import Command
from ...utils.functions import abstract


MAGIC_TYPES = {
    "cell": cell_magic,
    "line": line_magic,
    "line_cell": line_cell_magic
}


class IpythonCommandMagic(Command):
    """IPython Command base"""

    def __init__(self, magic, docstring, magic_type="cell"):
        self.__doc__ = docstring
        self.__name__ = type(self).__name__
        super(IpythonCommandMagic, self).__init__()
        self.is_ipython = True
        self.magic = magic
        self.docstring = docstring
        self.magic_type = magic_type
        self.args = []

    def add_argument_cmd(self, *args, **kwargs):
        """Ignore commands added by add_argument_cmd"""
        pass

    def add_argument(self, *args, **kwargs):
        """Add argument to parser available for both IPython magic and cmd"""
        self.args.append(
            magic_arguments.argument(*args, **kwargs)
        )

    def create_magic(self, func):
        """Create magic for command"""
        func.__name__ = str(self.magic)
        func.__doc__ = self.docstring
        func = MAGIC_TYPES[self.magic_type](self.magic)(func)
        for arg in self.args:
            func = arg(func)
        func = magic_arguments.magic_arguments()(func)
        return func

    def execute(self, func, line, cell, magic_cls):                              # pylint: disable=unused-argument, no-self-use, arguments-differ
        """Execute the command. Override on subclass"""
        abstract()

    def arguments(self, func, line):                                             # pylint: disable=no-self-use
        """Get arguments from magic"""
        argv = arg_split(line, posix=not sys.platform.startswith("win"))
        args = magic_arguments.parse_argstring(func, line)
        return argv, args



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Content Database"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import hashlib
import os

from os.path import join, isdir, isfile


CONTENT_DIRNAME = "content"


class ContentDatabase(object):
    """Content Database deal with storage of file content in disk"""

    def __init__(self, persistence_config):
        self.content_path = None  # Base path for storing content of files
        self.std_open = open  # Original Python open function.

        persistence_config.add(self)

    def set_path(self, config):
        """Set content_path"""
        self.content_path = join(config.provenance_path, CONTENT_DIRNAME)

    def mock(self, config):                                                      # pylint: disable=unused-argument, no-self-use
        """Mock storage for tests"""
        ContentDatabase.put = lambda s, c: hashlib.sha1(c).hexdigest()
        ContentDatabase.get = lambda s, c: "".encode("utf-8")

    def connect(self, config):
        """Create content directory"""
        if not config.should_mock and not isdir(self.content_path):
            os.makedirs(self.content_path)

    def put(self, content):
        """Put content in the content database

        Return: content hash code

        Arguments:
        content -- binary text to be saved
        """
        content_hash = hashlib.sha1(content).hexdigest()
        content_dirname = join(self.content_path, content_hash[:2])
        if not isdir(content_dirname):
            os.makedirs(content_dirname)
        content_filename = join(content_dirname, content_hash[2:])
        if not isfile(content_filename):
            with self.std_open(content_filename, "wb") as content_file:
                content_file.write(content)
        return content_hash

    def find_subhash(self, content_hash):
        """Get hash that starts by content_hash"""
        content_dirname = content_hash[:2]
        contet_filename = content_hash[2:]
        content_dir = join(self.content_path, content_dirname)
        if not isdir(content_dir):
            return None
        for _, _, filenames in os.walk(content_dir):
            for name in filenames:
                if name.startswith(contet_filename):
                    return content_dirname + name
        return None


    def get(self, content_hash):
        """Get content from the content database

        Return: content

        Arguments:
        content_hash -- content hash code
        """
        content_filename = join(self.content_path,
                                content_hash[:2],
                                content_hash[2:])
        with self.std_open(content_filename, "rb") as content_file:
            return content_file.read()



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Persistence Module. Perform database connection, and queries"""
from __future__ import (absolute_import, print_function,
                        division)

from .config import PersistenceConfig
from .content_database import ContentDatabase
from .relational_database import RelationalDatabase

persistence_config = PersistenceConfig()                                         # pylint: disable=invalid-name
content = ContentDatabase(persistence_config)                                    # pylint: disable=invalid-name
relational = RelationalDatabase(persistence_config)                              # pylint: disable=invalid-name


def get_serializer(arg):                                                         # pylint: disable=unused-argument
    """Select serializer according to argument"""
    # ToDo #54: use arg to select serialize
    # from .serializers import jsonpickle_serializer, jsonpickle_content
    # from .serializers import SimpleSerializer
    # return SimpleSerializer().serialize
    # return jsonpickle_serializer
    # return jsonpickle_content
    return repr


__all__ = [
    "persistence_config",
    "content",
    "relational",
    "get_serializer",
]



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""'now demo' command"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os
import subprocess
import sys

from ..utils.functions import resource, resource_is_dir, resource_ls

from .command import Command


DEMO = "../resources/demo"


def recursive_copy(origin, destiny):
    """Copy directory from resource to destiny folder"""
    if not os.path.exists(destiny):
        os.makedirs(destiny)
    for element in resource_ls(origin):
        origin_element = os.path.join(origin, element)
        destiny_element = os.path.join(destiny, element)
        if resource_is_dir(origin_element):
            recursive_copy(origin_element, destiny_element)
        else:
            with open(destiny_element, "wb") as fil:
                fil.write(resource(origin_element))


def erase(directory, everything=False):
    """Remove all files from directory


    Keyword Arguments:
    everything -- should delete .noworkflow too (default=False)
    """
    for root, dirs, files in os.walk(directory, topdown=False):
        for name in files:
            path = os.path.join(root, name)
            if everything or ".noworkflow" not in path:
                os.remove(os.path.join(root, name))
        for name in dirs:
            path = os.path.join(root, name)
            if everything or ".noworkflow" not in path:
                os.rmdir(path)


class Demo(Command):
    """Create demo project"""

    def add_arguments(self):
        add_arg = self.add_argument
        choices = sorted(list(resource_ls(DEMO)))
        add_arg("number", type=str, nargs="?", choices=choices,
                help="demo identification")
        add_arg("--steps", type=str, default=":",
                help="select step slice to run. Format start:stop:step")
        add_arg("--dir", type=str,
                help="set demo path. Default to CWD/demo<number>"
                     "where <number> is the demo identification")

    def execute(self, args):
        directory = "demo{}".format(args.number)
        if args.dir:
            directory = args.dir
        print("Creating Demo {}".format(args.number))
        demo_path = os.path.join(DEMO, args.number)
        steps = resource(os.path.join(demo_path, "steps.txt"),
                         encoding="utf-8")
        steps = steps.split("\n")
        steps = steps[slice(*[int(x) if x != "" else None
                              for x in args.steps.split(":")])]
        for step in steps:
            print(step)
            words = step.split(" ")
            param = " ".join(words[1:])
            if words[0] == ">LOAD":
                recursive_copy(os.path.join(demo_path, param),
                               directory)
            if words[0] == ">ERASE":
                erase(directory)
            if words[0] == ">ERASE_ALL":
                erase(directory, everything=True)
            if words[0] == "$now":
                cwd = os.getcwd()
                os.chdir(directory)
                subprocess.call("{} {}".format(sys.argv[0], param),
                                shell=True, stdout=subprocess.PIPE)
                os.chdir(cwd)
        os.chdir(directory)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Trial Dot Object"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import weakref

from future.utils import viewitems

from .base import Model
from . import FileAccess
from .graphs.dependency_graph import DependencyFilter, variable_id
from .graphs.dependency_graph import DotVisitor


CALL_SCHEMA = "#3A85B9", "box", "black"
VAR_SCHEMA = "#85CBD0", "ellipse", "black"
FILE_SCHEMA = "white", "ellipse", "black"
BLACKBOX_SCHEMA = "black", "box", "grey"
GRAYBOX_SCHEMA = "grey", "box", "black"
IMPORT_SCHEMA = "#1B2881", "box", "#7AC5F9"


TYPES = {
    "call": CALL_SCHEMA,
    "normal": VAR_SCHEMA,
    "virtual": VAR_SCHEMA,
    "param": VAR_SCHEMA,
    "import": IMPORT_SCHEMA,
    "--blackbox--": BLACKBOX_SCHEMA,
    "--graybox--": GRAYBOX_SCHEMA,
    "access": FILE_SCHEMA,
}


class TrialDot(Model):                                                           # pylint: disable=too-many-instance-attributes
    """Handle Dot export"""

    __modelname__ = "TrialDot"

    def __init__(self, trial):
        super(TrialDot, self).__init__()
        self.trial = weakref.proxy(trial)
        self.format = "svg"
        self.value_length = 0
        self.name_length = 55
        self.fallback = None
        self.run = True


    def simulation(self):
        """Configure simulation graph"""
        self.fallback = None

    def prospective(self):
        """Configure prospective graph"""
        self.fallback = None

    def dependency(self):
        """Configure dependency graph"""
        self.fallback = VAR_SCHEMA

    def export_text(self):
        """Export facts from trial as text"""
        dep_filter = self.trial.dependency_filter
        if self.run:
            dep_filter.run()
        getattr(self, self.trial.dependency_config.mode)()
        visitor = DotVisitor(self.fallback, self.name_length,
                             self.value_length, TYPES, dep_filter)
        visitor.visit(dep_filter.main_cluster)
        return "\n".join(visitor.result)

    def _repr_svg_(self):
        if self.format == "svg":
            ipython = get_ipython()
            return ipython.run_cell_magic(
                "dot", "--format {}".format(self.format), self.export_text()
            )

    def _repr_png_(self):
        if self.format == "png":
            ipython = get_ipython()
            return ipython.run_cell_magic(
                "dot", "--format {}".format(self.format), self.export_text()
            )



#!/usr/bin/python2

import csv
from collections import defaultdict


def read(filename):
    result = defaultdict(list)
    with open(filename, 'r') as c:
        reader = csv.reader(c, delimiter=";")
        for row in reader:
            month = int(row[1].split('/')[1])
            precipitation = float(row[3])
            result[month].append(precipitation)
    return result


def prepare(series, months, names, plt, div=.1, colors=['b', 'g', 'r']):
    fig, ax = plt.subplots()

    ax.set_ylabel('Precipitation (mm)')
    ax.set_xlabel('Month')
    ax.set_title('Precipitation by Month')
    ax.set_xticks(months + .5)
    ax.set_xticklabels(list(map(str, months)))

    half_div = div / 2.0
    width = (1.0 - div) / len(series)
    bars = []
    for i, data in enumerate(series):
        bars.append(ax.bar(months + half_div + i * width, data, width,
                           color=colors[i]))

    ax.legend(bars, names)

__VERSION__ = "1.0.1"

if __name__ == '__main__':
    import sys
    filename = sys.argv[1]
    month = sys.argv[2]
    data = read(filename)
    print(';'.join(map(str, data[int(month)])))



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""'now debug' command"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from .cmd_run import Run


class Debug(Run):
    """Debug a script collecting its provenance
       This command is similar to Run, but defines different variables"""

    def __init__(self, *args, **kwargs):
        super(Debug, self).__init__(*args, **kwargs)
        self.default_context = "package"
        self.default_call_storage_frequency = 1
        self.default_save_frequency = 1000
        self.default_execution_provenance = "Tracer"



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Profiler Provider"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)


import sys
import os
import traceback
import time

from datetime import datetime

from ...persistence import content
from ...persistence.models import Activation, ObjectValue, FileAccess, Trial
from ...utils.cross_version import builtins

from .base import ExecutionProvider
from .argument_captors import ProfilerArgumentCaptor


class Profiler(ExecutionProvider):                                               # pylint: disable=too-many-instance-attributes
    """Profiler

    Collect:
        activations, arguments, returns, global variables, and file accesses
    """

    def __init__(self, *args):
        super(Profiler, self).__init__(*args)
        # Open
        content.std_open = open
        builtins.open = self.new_open(open)

        # the number of user functions activated
        #   (starts with -1 to compensate the first call to the script itself)
        self.depth_user = -1
        # the number of non-user functions activated
        self.depth_non_user = 0
        # The first caller is None
        self.activation_stack = [None]

        # Store provenance
        self.activations = self.metascript.activations_store
        self.object_values = self.metascript.object_values_store
        self.file_accesses = self.metascript.file_accesses_store

        # Avoid using the same event for tracer and profiler
        self.last_event = None

        self.definition = self.metascript.definition
        self.function_globals = self.definition.function_globals

        self.event_map["c_call"] = self.trace_c_call
        self.event_map["call"] = self.trace_call
        self.event_map["c_return"] = self.trace_c_return
        self.event_map["c_exception"] = self.trace_c_exception
        self.event_map["return"] = self.trace_return

        # Partial save
        self.save_frequency = self.metascript.save_frequency / 1000.0
        self.call_storage_frequency = self.metascript.call_storage_frequency
        self.closed_activations = 0

        self.timer = time.time
        self.last_time = self.timer()

        # Events are unique
        self.unique_events = True

        # Global activations
        self.globals = {}
        # Main activation
        self.main_activation = None

        # Skip tear_up return
        self.skip_first_return = True
        self.enabled = True

        # Capture arguments
        self.argument_captor = ProfilerArgumentCaptor(self)

        if sys.version_info >= (3, 0):
            Profiler.add_activation = Profiler.initial_add_activation

    @property
    def current_activation(self):
        """Return current activation"""
        astack = self.activation_stack
        if astack[-1] is not None:
            return self.activations[astack[-1]]
        return self.activations.dry_add("", "", "empty", 0, 0, -1, False)

    @property
    def parent_activation(self):
        """Return activation that called current activation"""
        return self.activations[self.activation_stack[-2]]

    def new_open(self, old_open):
        """Wrap the open builtin function to register file access"""
        def open(name, *args, **kwargs):                                         # pylint: disable=redefined-builtin
            """Open file and add it to file_accesses"""
            if self.enabled:
                # Create a file access object with default values
                fid = self.file_accesses.add(name)
                file_access = self.file_accesses[fid]

                if os.path.exists(name):
                    # Read previous content if file exists
                    with old_open(name, "rb") as fil:
                        file_access.content_hash_before = content.put(
                            fil.read()
                        )

                # Update with the informed keyword arguments (mode / buffering)
                file_access.update(kwargs)
                # Update with the informed positional arguments
                if len(args) > 1:
                    file_access.buffering = args[1]
                elif len(args) > 0:
                    file_access.mode = args[0]

                self.add_file_access(file_access)
            return old_open(name, *args, **kwargs)

        return open

    def add_file_access(self, file_access):
        """After activation that called open finish, add file_accesses to it"""
        activation = self.current_activation
        file_access.function_activation_id = activation.id
        activation.file_accesses.append(file_access)

    def valid_depth(self, extra=0):
        """Check if it is capturing in a valid depth
        Consider both user depth and non user depth.
        """
        depth = self.depth_user + self.depth_non_user + extra
        if depth < 0:
            self.enabled = False
            return False
        if depth > self.depth_threshold:
            return False
        return self.depth_non_user <= self.non_user_depth_threshold

    def add_activation(self, aid):                                               # pylint: disable=function-redefined
        """Add activation to activation stack"""
        self.activation_stack.append(aid)

    _old_add_activation = add_activation

    def initial_add_activation(self, aid):
        """Add activation to activation stack.
        Ignore first module.exec on Python 3"""
        Profiler.add_activation = Profiler._old_add_activation
        if aid == 1 and self.activations[aid].name == "module.exec":
            self.depth_non_user -= 1
            self.activations.id = 0
            return
        return Profiler._old_add_activation(self, aid)

    def close_activation(self, frame, event, arg):                               # pylint: disable=unused-argument
        """Remove activation from stack, set finish time and add accesses"""
        activation = self.current_activation
        self.activation_stack.pop()
        activation.finish = datetime.now()
        try:
            if event == "return":
                activation.return_value = self.serialize(arg)
        except Exception:  # ignoring any exception during capture               # pylint: disable=broad-except
            activation.return_value = None
        # Update content of accessed files
        for file_access in activation.file_accesses:
            # Checks if file still exists
            if os.path.exists(file_access.name):
                with content.std_open(file_access.name, "rb") as fil:
                    file_access.content_hash_after = content.put(fil.read())
            file_access.done = True
        self.closed_activations += 1
        if (self.call_storage_frequency and
                (self.closed_activations % self.call_storage_frequency == 0)):
            self.store(partial=True)

    def trace_c_call(self, frame, event, arg):                                   # pylint: disable=unused-argument
        """Trace c_call. Increase non_user depth"""
        self.depth_non_user += 1
        if self.valid_depth():
            self.add_activation(self.activations.add(
                "now(n/a)",
                frame.f_code.co_filename,
                arg.__name__ if arg.__self__ is None else ".".join(
                    [type(arg.__self__).__name__, arg.__name__]),
                frame.f_lineno, frame.f_lasti, self.activation_stack[-1],
                False
            ))

    def trace_call(self, frame, event, arg):                                     # pylint: disable=unused-argument
        """Trace call. Increase depth and create activation"""
        co_name = frame.f_code.co_name
        co_filename = frame.f_code.co_filename
        if co_filename in self.paths:
            self.depth_user += 1
            in_paths = True
        else:
            self.depth_non_user += 1
            in_paths = False

        if self.valid_depth():
            aid = self.activations.add(
                co_filename,
                frame.f_back.f_code.co_filename,
                co_name if co_name != "<module>" else co_filename,
                frame.f_back.f_lineno, frame.f_back.f_lasti,
                self.activation_stack[-1],
                in_paths and self.valid_depth(extra=1)
            )
            activation = self.activations[aid]

            if activation.is_main:
                self.main_activation = activation
            if co_name == "<module>":
                self.globals[co_filename] = activation
            # Capturing arguments
            self.argument_captor.capture(frame, activation)

            # Capturing globals
            def_globals = self.function_globals[co_filename][activation.name]
            fglobals = frame.f_globals
            for global_var in def_globals:
                self.object_values.add(
                    global_var, self.serialize(fglobals[global_var]),
                    "GLOBAL", aid)

            activation.start = datetime.now()
            self.add_activation(aid)

    def trace_c_return(self, frame, event, arg):                                 # pylint: disable=unused-argument
        """Trace c_return. Decrease non_user depth"""
        if self.valid_depth():
            self.close_activation(frame, event, arg)
        self.depth_non_user -= 1

    def trace_c_exception(self, frame, event, arg):
        """Trace c_exception. Decrease non_user depth"""
        if self.valid_depth():
            self.close_activation(frame, event, arg)
        self.depth_non_user -= 1

    def trace_return(self, frame, event, arg):
        """Trace return. Decrease depth and close activation"""
        # Only enable activations gathering after the first call to the script
        if self.skip_first_return:
            self.skip_first_return = False
            return
        if self.valid_depth():
            self.close_activation(frame, event, arg)

        if frame.f_code.co_filename in self.paths:
            if frame.f_code.co_name == "<module>":
                self.enabled = False
            self.depth_user -= 1
        else:
            self.depth_non_user -= 1

    def new_event(self, frame, event, arg):                                      # pylint: disable=unused-argument
        """Check if event is new to avoid Profiler and Tracer overlap"""
        current_event = (event, frame.f_lineno, id(frame))
        if self.last_event != current_event:
            self.last_event = current_event
            return True
        return False

    def tracer(self, frame, event, arg):
        """Check if event is valid before executing tracer
        Call event function from event_map
        """
        try:
            if self.enabled:
                if self.unique_events or self.new_event(frame, event, arg):
                    self.pre_tracer(frame, event, arg)
                    self.event_map[event](frame, event, arg)
                if (self.save_frequency and
                        (self.timer() - self.last_time > self.save_frequency)):
                    self.store(partial=True)
                    self.last_time = self.timer()
        except Exception:                                                        # pylint: disable=broad-except
            traceback.print_exc()
        finally:
            return self.tracer                                                   # pylint: disable=lost-exception

    def pre_tracer(self, frame, event, arg):                                     # pylint: disable=unused-argument, no-self-use
        """It is executed before the tracing event"""
        pass

    def store(self, partial=False):
        """Store execution provenance"""
        tid = self.trial_id
        if not partial:
            now = datetime.now()
            Trial.fast_update(tid, now, self.metascript.docstring)

        Activation.fast_store(tid, self.activations, partial)
        ObjectValue.fast_store(tid, self.object_values, partial)
        FileAccess.fast_store(tid, self.file_accesses, partial)

    def tearup(self):
        """Activate profiler"""
        sys.setprofile(self.tracer)

    def teardown(self):
        """Deactivate profiler"""
        builtins.open = content.std_open
        super(Profiler, self).teardown()
        sys.setprofile(self.default_profile)



# Copyright (c) 2014 Universidade Federal Fluminense (UFF)
# Copyright (c) 2014 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from .call_slicing_test import TestCallSlicing

__all__ = [
    b'TestCallSlicing',
]



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""'now show' command"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os

from ..ipython.converter import create_ipynb
from ..persistence.models import Trial
from ..persistence import persistence_config
from ..utils.functions import wrap
from ..utils.io import print_msg

from .command import NotebookCommand


def print_trial_relationship(relation, breakline="\n\n", other="\n    "):
    """Print trial relationship"""
    output = []
    for obj in relation:
        obj.show(_print=lambda x: output.append(wrap(x, other=other)))
    print(breakline.join(output))


def print_function_activation(trial, activation, level=1):
    """Print function activation recursively"""
    text = wrap(
        "{0.line}: {0.name} ({0.start} - {0.finish})".format(activation),
        initial="  " * level)
    indent = text.index(": ") + 2
    print(text)
    activation.show(_print=lambda x, offset=0: print(
        wrap(x, initial=" " * (indent + offset))))

    for inner_activation in activation.children:
        print_function_activation(trial, inner_activation, level + 1)


class Show(NotebookCommand):
    """Show the collected provenance of a trial"""

    def add_arguments(self):
        add_arg = self.add_argument
        add_arg("trial", type=str, nargs="?",
                help="trial id or none for last trial")
        add_arg("-m", "--modules", action="store_true",
                help="shows module dependencies")
        add_arg("-d", "--function-defs", action="store_true",
                help="shows the user-defined functions")
        add_arg("-e", "--environment", action="store_true",
                help="shows the environment conditions")
        add_arg("-a", "--function-activations", action="store_true",
                help="shows function activations")
        add_arg("-f", "--file-accesses", action="store_true",
                help="shows read/write access to files")
        add_arg("--dir", type=str,
                help="set project path where is the database. Default to "
                     "current directory")

    def execute(self, args):
        persistence_config.connect_existing(args.dir or os.getcwd())
        trial = Trial(trial_ref=args.trial)

        print_msg("trial information:", True)
        trial.show(_print=lambda x: print(wrap(x)))

        if args.modules:
            print_msg("this trial depends on the following modules:", True)
            print_trial_relationship(trial.modules)

        if args.function_defs:
            print_msg("this trial has the following functions:", True)
            print_trial_relationship(trial.function_defs)

        if args.environment:
            print_msg("this trial has been executed under the following"
                      " environment conditions", True)
            print_trial_relationship(trial.environment_attrs, breakline="\n",
                                     other="\n  ")

        if args.function_activations:
            print_msg("this trial has the following function activation "
                      "graphF:", True)
            for act in trial.initial_activations:                                # pylint: disable=not-an-iterable
                print_function_activation(trial, act)

        if args.file_accesses:
            print_msg("this trial accessed the following files:", True)
            print_trial_relationship(trial.file_accesses)

    def execute_export(self, args):
        persistence_config.connect_existing(args.dir or os.getcwd())
        Trial(trial_ref=args.trial)
        name = "Trial-{}.ipynb".format(args.trial)
        code = ("%load_ext noworkflow\n"
                "import noworkflow.now.ipython as nip\n"
                "# <codecell>\n"
                "trial = nip.Trial({})\n"
                "# trial.graph.mode = 3\n"
                "# trial.graph.width = 500\n"
                "# trial.graph.height = 500\n"
                "# <codecell>\n"
                "trial").format(repr(args.trial) if args.trial else '')
        if not args.trial:
            name = "Current Trial.ipynb"
            code += (
                "\n"
                "# <codecell>\n"
                "%now_ls_magic\n"
                "# <codecell>\n"
                "%%now_prolog {trial.id}\n"
                "activation({trial.id}, 1, X, _, _, _)"
            )

        create_ipynb(name, code)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Define io utility functions"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import sys

from future.utils import viewitems

from .cross_version import StringIO


LABEL = "[now] "
verbose = False                                                                  # pylint: disable=invalid-name

STDIN = sys.stdin
STDOUT = sys.stdout
STDERR = sys.stderr


class redirect_output(object):                                                   # pylint: disable=invalid-name, too-few-public-methods
    """Redirect output to stream"""

    def __init__(self, outputs=None):
        if outputs is None:
            outputs = ["stdout", "stderr"]
        self.outputs = outputs
        self.old = {}

    def __enter__(self):
        result = []
        for out in self.outputs:
            self.old[out] = getattr(sys, out)
            setattr(sys, out, StringIO())
            result.append(self.old[out])
        return result

    def __exit__(self, exc_type, value, traceback):
        for out, old in viewitems(self.old):
            setattr(sys, out, old)


def print_msg(message, force=False, file=STDOUT):
    """Print message with [now] prefix when in verbose mode"""
    if verbose or force:
        print("{}{}".format(LABEL, message), file=file)


def print_fn_msg(message, force=False, file=STDOUT):
    """Print lazy message with [now] prefix"""
    if verbose or force:
        print("{}{}".format(LABEL, message()), file=file)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
""""now list" command"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os

from ..persistence.models import Trial
from ..persistence import persistence_config
from ..utils.io import print_msg

from .command import Command


class List(Command):
    """List all trials registered in the current directory"""

    def add_arguments(self):
        add_arg = self.add_argument
        add_arg("--dir", type=str,
                help="set project path where is the database. Default to "
                     "current directory")

    def execute(self, args):
        persistence_config.connect_existing(args.dir or os.getcwd())
        print_msg("trials available in the provenance store:", True)
        for trial in Trial.all():
            text = "  Trial {0.id}: {0.command}".format(trial)
            indent = text.index(": ") + 2
            print(text)
            print("{indent}with code hash {t.code_hash}".format(
                indent=" " * indent, t=trial))
            print("{indent}ran from {t.start} to {t.finish}".format(
                indent=" " * indent, t=trial))
            if trial.finish:
                print('{indent}duration: {t.duration_text}'.format(
                    indent=" " * indent, t=trial))



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Capture arguments from calls"""

from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import weakref
import itertools
import inspect

from future.utils import viewitems

from ...utils.functions import abstract

from ..prov_definition.utils import ClassDef, Assert, With, Decorator


WITHOUT_PARAMS = (ClassDef, Assert, With)


class ArgumentCaptor(object):                                                    # pylint: disable=too-few-public-methods
    """Collect arguments during calls"""

    def __init__(self, provider):
        self.provider = weakref.proxy(provider)

    def capture(self, frame, activation):                                        # pylint: disable=unused-argument, no-self-use
        """Abstract method for capture"""
        abstract()


class ProfilerArgumentCaptor(ArgumentCaptor):                                    # pylint: disable=too-few-public-methods
    """Collect arguments for profiler"""

    def __init__(self, *args, **kwargs):
        super(ProfilerArgumentCaptor, self).__init__(*args, **kwargs)
        self.f_locals = {}

    def capture(self, frame, activation):
        """Store argument object values


        Arguments:
        frame -- current frame, after trace call
        activation -- current activation
        """
        provider = self.provider
        self.f_locals = values = frame.f_locals
        code = frame.f_code
        names = code.co_varnames
        nargs = code.co_argcount
        # Capture args
        for var in itertools.islice(names, 0, nargs):
            try:
                provider.object_values.add(
                    var,
                    provider.serialize(values[var]), "ARGUMENT", activation.id)
                activation.args.append(var)
            except Exception:                                                    # pylint: disable=broad-except
                # ignoring any exception during capture
                pass
        # Capture *args
        if code.co_flags & inspect.CO_VARARGS:                                   # pylint: disable=no-member
            varargs = names[nargs]
            provider.object_values.add(
                varargs,
                provider.serialize(values[varargs]), "ARGUMENT", activation.id)
            activation.starargs.append(varargs)
            nargs += 1
        # Capture **kwargs
        if code.co_flags & inspect.CO_VARKEYWORDS:                               # pylint: disable=no-member
            kwargs = values[names[nargs]]
            for key in kwargs:
                provider.object_values.add(
                    key, provider.serialize(kwargs[key]), "ARGUMENT",
                    activation.id)
            activation.kwargs.append(names[nargs])


class InspectProfilerArgumentCaptor(ArgumentCaptor):                             # pylint: disable=too-few-public-methods
    """This Argument Captor uses the inspect.getargvalues that is slower
    because it considers the existence of anonymous tuple
    """

    def capture(self, frame, activation):
        """Store argument object values


        Arguments:
        frame -- current frame, after trace call
        activation -- current activation
        """
        provider = self.provider
        # ToDo #75: inspect.getargvalues was deprecated on Python 3.5
        # ToDo #75: use inspect.signature instead
        (args, varargs, keywords, values) = inspect.getargvalues(frame)
        for arg in args:
            try:
                provider.object_values.add(
                    arg, provider.serialize(values[arg]), "ARGUMENT",
                    activation.id)
                activation.args.append(arg)
            except Exception:  # ignoring any exception during capture           # pylint: disable=broad-except
                pass
        if varargs:
            provider.object_values.add(
                varargs, provider.serialize(values[varargs]), "ARGUMENT",
                activation.id)
            activation.starargs.append(varargs)
        if keywords:
            for key, value in viewitems(values[keywords]):
                provider.object_values.add(
                    key, provider.serialize(value), "ARGUMENT", activation.id)
                activation.kwargs.append(key)


class SlicingArgumentCaptor(ProfilerArgumentCaptor):
    """Create Slicing Variables for Arguments and dependencies between
    Parameters and Arguments"""


    def __init__(self, *args, **kwargs):
        super(SlicingArgumentCaptor, self).__init__(*args, **kwargs)
        self.caller, self.activation = None, None
        self.filename, self.line = "", 0
        self.frame = None

    def match_arg(self, passed, arg):
        """Match passed arguments with param


        Arguments:
        passed -- Call Variable name
        arg -- Argument name
        """
        provider = self.provider
        activation = self.activation
        context = activation.context

        if arg in context:
            act_var = context[arg]
        else:
            vid = provider.add_variable(activation.id, arg,
                                        self.line, self.f_locals, "param")
            act_var = provider.variables[vid]
            context[arg] = act_var

        if passed:
            caller = self.caller
            target = provider.find_variable(caller, passed, self.filename)
            if target is not None:
                provider.dependencies.add(
                    act_var.activation_id, act_var.id,
                    target.activation_id, target.id, "parameter"
                )

    def match_args(self, params, arg):
        """Match passed argument with param


        Arguments:
        params -- Call Variable names
        arg -- Argument name
        """
        for param in params:
            self.match_arg(param, arg)

    def _defined_call(self, activation):
        """Return a call extracted from AST if it has arguments
        or None, otherwise


        Arguments:
        activation -- current activation
        """
        if not activation.with_definition or activation.is_main:
            return
        if activation.is_comprehension():
            return
        provider = self.provider
        lineno, lasti = activation.line, activation.lasti
        filename = activation.filename
        function_name = activation.name
        if (function_name == "__enter__" and
                lasti in provider.with_enter_by_lasti[filename][lineno]):
            activation.has_parameters = False
            return
        if (function_name == "__exit__" and
                lasti in provider.with_exit_by_lasti[filename][lineno]):
            activation.has_parameters = False
            return
        if lasti in provider.iters[filename][lineno]:
            activation.has_parameters = False
            provider.next_is_iter = True
            return
        try:
            call = provider.call_by_lasti[filename][lineno][lasti]
        except (IndexError, KeyError):
            # call not found
            # ToDo: show in dev-mode
            return
        if (isinstance(call, WITHOUT_PARAMS) or
                (isinstance(call, Decorator) and not call.is_fn)):
            activation.has_parameters = False
            return

        return call

    def capture(self, frame, activation):                                        # pylint: disable=too-many-locals
        """Match call parameters to function arguments


        Arguments:
        frame -- current frame, after trace call
        activation -- current activation
        """
        super(SlicingArgumentCaptor, self).capture(frame, activation)
        provider = self.provider

        self.frame = frame
        call = self._defined_call(activation)
        if not call:
            return
        self.filename = activation.filename

        self.line = frame.f_lineno
        self.caller, self.activation = provider.current_activation, activation

        match_args, match_arg = self.match_args, self.match_arg
        act_args_index = activation.args.index

        # Check if it has starargs and kwargs
        sub = -[bool(activation.starargs), bool(activation.kwargs)].count(True)
        order = activation.args + activation.starargs + activation.kwargs
        activation_arguments = len(order) + sub
        used = [0 for _ in order]
        j = 0

        # Match positional arguments
        for i, call_arg in enumerate(call.args):
            if call_arg:
                j = i if i < activation_arguments else sub
                act_arg = order[j]
                match_args(call_arg, act_arg)
                used[j] += 1

        # Match keyword arguments
        for act_arg, call_arg in viewitems(call.keywords):
            try:
                i = act_args_index(act_arg)
                match_args(call_arg, act_arg)
                used[i] += 1
            except ValueError:
                for kwargs in activation.kwargs:
                    match_args(call_arg, kwargs)

        # Match kwargs, starargs
        # ToDo #75: Python 3.5 supports multiple keyword arguments and starargs
        # ToDo #75: improve matching
        #   Ignore default params
        #   Do not match f(**kwargs) with def(*args)
        args = [(k, order[k]) for k in range(len(used)) if not used[k]]
        for star in call.kwargs + call.starargs:
            for i, act_arg in args:
                match_args(star, act_arg)
                used[i] += 1

        # Create variables for unmatched arguments
        args = [(k, order[k]) for k in range(len(used)) if not used[k]]
        for i, act_arg in args:
            match_arg(None, act_arg)

        # Create dependencies between all parameters
        # ToDo #35: improve dependencies to use references.
        #   Do not create dependencies between all parameters
        all_args = list(provider.find_variables(
            self.caller, call.all_args(), activation.filename))
        if all_args:
            graybox = provider.create_graybox()
            provider.add_dependencies(graybox, all_args)
            provider.add_inter_dependencies(frame.f_back.f_locals, all_args,
                                            self.caller, activation.line,
                                            [(graybox, graybox.name)])
